'use strict';

const
	logger = require('./lib/logger');

const
	_ = require('lodash'),
	path = require('path'),
	http = require('http'),
	fs = require('fs-extra'),
	express = require('express'),
	highland = require('highland'),
	Promise = require('bluebird'),
	config = require('./lib/config'),
	cache = require('./lib/cache'),
	shared = require('./lib/shared'),
	TestCase = require('./lib/test-case'),
	SVGStream = require('./lib/svg-stream'),
	winston = require('winston');

const
	app = express(),
	server = http.createServer(app);

// app config
app.set('x-powered-by', false);

// static files
app.use('/public', express.static(path.join(__dirname, 'public')));

// service endpoint
app.get('/:slug/:rev?', (req, res, next) => {
	const query = req.query;
	const params = req.params;
	const client = cache.getClient();
	if (/\.[a-z0-9]+$/i.test(params.slug)) {
		// continue if this is a request for a static asset
		next();
	} else {
		let testCase = new TestCase(params.slug, params.rev);
		testCase
			.fetch()
			.then(() => {
				switch (query.o) {
					case 'json':
						return res.json(testCase);
					default:
						return cache
							.expired(client, testCase.imageKey)
							.then(() => {
								req.url = testCase.image;
								next();
							});
				}
			})
			.catch((err) => {
				if (err.code != 'ENOTCACHED') {
					throw err;
				}
				return SVGStream(testCase.link)
					.pipe(highland())
					.through(shared.streamToPromise)
					.then((bufferArray) => {
						let buffer = shared.joinBuffer(bufferArray);
						res.setHeader('Content-Type', config.mime);
						res.end(buffer);
						return cache.write(client, testCase.imageKey, buffer.toString('base64'), config.imageTTL);
					});
			})
			.then(() => logger.info(`200 ${req.originalUrl}`))
			.timeout(config.timeout)
			.catch(next);
	}
});

// cache endpoint
app.get('/cache/:prefix/:tag', (req, res, next) => {
	const params = req.params;
	const client = cache.getClient();
	const key = `${params.prefix}:${params.tag}`;
	res.setHeader('Content-Type', config.mime);
	return cache
		.read(client, key)
		.then(str => res.end(new Buffer(str, 'base64')))
		.catch(next);
});

// handle timeout
app.use((err, req, res, next) => {
	if (err instanceof Promise.TimeoutError) {
		err.status = 408;
	}
	next(err);
});

// catch 404 and forward to error handler
app.use((req, res, next) => {
	var err = new Error('Not Found');
	err.status = 404;
	next(err);
});

// handle errors
app.use((err, req, res, next) => {
	if (_.get(err, 'status', 500) !== 500) {
		logger.warn(`${err.status} | ${err.message} | ${req.originalUrl}`);
	} else {
		logger.error(err);
	}

	res.status(err.status || 500);

	if (!res.headersSent) {
		res.setHeader('Content-Type', 'text/plain');
		res.json({error: err.message});
	} else {
		res.end();
	}
});

server.listen(config.port, () => {
	logger.info(`jsperf service listening at ${config.localBase}.`);
});
